@@ -18,7 +18,7 @@ import geomet.wkt
 import html2text
 import requests
 from six import string_types
-from six.moves.urllib.parse import urljoin
+from six.moves.urllib.parse import urljoin, quote_plus
 from tqdm import tqdm
 
 from . import __version__ as sentinelsat_version
@@ -31,29 +31,36 @@ class SentinelAPI:
     ----------
     user : string
         username for DataHub
+        set to None to use ~/.netrc
     password : string
         password for DataHub
+        set to None to use ~/.netrc
     api_url : string, optional
         URL of the DataHub
         defaults to 'https://scihub.copernicus.eu/apihub'
     show_progressbars : bool
         Whether progressbars should be shown or not, e.g. during download. Defaults to True.
+    timeout : float or tuple, optional
+        How long to wait for DataHub response (in seconds).
+        Tuple (connect, read) allowed.
 
     Attributes
     ----------
-    session : requests.Session object
+    session : requests.Session
         Session to connect to DataHub
     api_url : str
         URL to the DataHub
     page_size : int
-        number of results per query page
-        current value: 100 (maximum allowed on ApiHub)
+        Number of results per query page.
+        Current value: 100 (maximum allowed on ApiHub)
+    timeout : float or tuple
+        How long to wait for DataHub response (in seconds).
     """
 
     logger = logging.getLogger('sentinelsat.SentinelAPI')
 
     def __init__(self, user, password, api_url='https://scihub.copernicus.eu/apihub/',
-                 show_progressbars=True):
+                 show_progressbars=True, timeout=None):
         self.session = requests.Session()
         if user and password:
             self.session.auth = (user, password)
@@ -62,6 +69,7 @@ class SentinelAPI:
         self.user_agent = 'sentinelsat/' + sentinelsat_version
         self.session.headers['User-Agent'] = self.user_agent
         self.show_progressbars = show_progressbars
+        self.timeout = timeout
         # For unit tests
         self._last_query = None
         self._last_response = None
@@ -80,6 +88,7 @@ class SentinelAPI:
             Expects a tuple of (start, end), e.g. ("NOW-1DAY", "NOW").
             The timestamps can be either a Python datetime or a string in one of the
             following formats:
+
                 - yyyyMMdd
                 - yyyy-MM-ddThh:mm:ss.SSSZ (ISO-8601)
                 - yyyy-MM-ddThh:mm:ssZ
@@ -88,35 +97,39 @@ class SentinelAPI:
                 - NOW+<n>DAY(S)
                 - yyyy-MM-ddThh:mm:ssZ-<n>DAY(S)
                 - NOW/DAY (or HOUR, MONTH etc.) - rounds the value to the given unit
+
             Alternatively, an already fully formatted string such as "[NOW-1DAY TO NOW]" can be
             used as well.
         raw : str, optional
             Additional query text that will be appended to the query.
-        area_relation : {'Intersection', 'Contains', 'IsWithin'}, optional
+        area_relation : {'Intersects', 'Contains', 'IsWithin'}, optional
             What relation to use for testing the AOI. Case insensitive.
+
                 - Intersects: true if the AOI and the footprint intersect (default)
                 - Contains: true if the AOI is inside the footprint
                 - IsWithin: true if the footprint is inside the AOI
+
         order_by: str, optional
             A comma-separated list of fields to order by (on server side).
             Prefix the field name by '+' or '-' to sort in ascending or descending order,
-            respectively. Ascending order is used, if prefix is omitted.
+            respectively. Ascending order is used if prefix is omitted.
             Example: "cloudcoverpercentage, -beginposition".
         limit: int, optional
             Maximum number of products returned. Defaults to no limit.
         offset: int, optional
             The number of results to skip. Defaults to 0.
+        **keywords
+            Additional keywords can be used to specify other query parameters,
+            e.g. `relativeorbitnumber=70`.
+            See https://scihub.copernicus.eu/twiki/do/view/SciHubUserGuide/3FullTextSearch
+            for a full list.
 
-        Other Parameters
-        ----------------
-        Additional keywords can be used to specify other query parameters,
-        e.g. relativeorbitnumber=70.
-        See https://scihub.copernicus.eu/twiki/do/view/SciHubUserGuide/3FullTextSearch
-        for a full list.
 
-        Range values can be passed as two-element tuples, e.g. cloudcoverpercentage=(0, 30).
+        Range values can be passed as two-element tuples, e.g. `cloudcoverpercentage=(0, 30)`.
+        `None` can be used in range values for one-sided ranges, e.g. `orbitnumber=(16302, None)`.
+        Ranges with no bounds (`orbitnumber=(None, None)`) will not be included in the query.
 
-        The time interval formats accepted by the ``date`` parameter can also be used with
+        The time interval formats accepted by the `date` parameter can also be used with
         any other parameters that expect time intervals (that is: 'beginposition', 'endposition',
         'date', 'creationdate', and 'ingestiondate').
 
@@ -138,18 +151,32 @@ class SentinelAPI:
     @staticmethod
     def format_query(area=None, date=None, raw=None, area_relation='Intersects',
                      **keywords):
-        """Create OpenSearch API query string
+        """Create a OpenSearch API query string.
         """
         if area_relation.lower() not in {"intersects", "contains", "iswithin"}:
             raise ValueError("Incorrect AOI relation provided ({})".format(area_relation))
 
+        # Check for duplicate keywords
+        kw_lower = set(x.lower() for x in keywords)
+        if (len(kw_lower) != len(keywords) or
+                (date is not None and 'beginposition' in kw_lower) or
+                (area is not None and 'footprint' in kw_lower)):
+            raise ValueError("Query contains duplicate keywords. Note that query keywords are case-insensitive.")
+
         query_parts = []
 
         if date is not None:
             keywords['beginPosition'] = date
 
         for attr, value in sorted(keywords.items()):
-            # From https://github.com/SentinelDataHub/DataHubSystem/search?q=text/date+iso8601
+            # Escape spaces, where appropriate
+            if isinstance(value, string_types):
+                value = value.strip()
+                if not any(value.startswith(s[0]) and value.endswith(s[1]) for s in ['[]', '{}', '//', '()']):
+                    value = re.sub(r'\s', r'\ ', value, re.M)
+
+            # Handle date keywords
+            # Keywords from https://github.com/SentinelDataHub/DataHubSystem/search?q=text/date+iso8601
             date_attrs = ['beginposition', 'endposition', 'date', 'creationdate', 'ingestiondate']
             if attr.lower() in date_attrs:
                 # Automatically format date-type attributes
@@ -163,9 +190,14 @@ class SentinelAPI:
                     raise ValueError("Date-type query parameter '{}' expects a two-element tuple "
                                      "of str or datetime objects. Received {}".format(attr, value))
 
+            # Handle ranged values
             if isinstance(value, (list, tuple)):
                 # Handle value ranges
                 if len(value) == 2:
+                    # Allow None to be used as a unlimited bound
+                    value = ['*' if x is None else x for x in value]
+                    if all(x == '*' for x in value):
+                        continue
                     value = '[{} TO {}]'.format(*value)
                 else:
                     raise ValueError("Invalid number of elements in list. Expected 2, received "
@@ -179,14 +211,14 @@ class SentinelAPI:
         if area is not None:
             query_parts.append('footprint:"{}({})"'.format(area_relation, area))
 
-        query = ' '.join(query_parts)
-        return query
+        return ' '.join(query_parts)
 
     def query_raw(self, query, order_by=None, limit=None, offset=0):
-        """Do a full-text query on the OpenSearch API using the format specified in
-           https://scihub.copernicus.eu/twiki/do/view/SciHubUserGuide/3FullTextSearch
+        """
+        Do a full-text query on the OpenSearch API using the format specified in
+        https://scihub.copernicus.eu/twiki/do/view/SciHubUserGuide/3FullTextSearch
 
-        DEPRECATED: use query(raw=...) instead. This method will be removed in the next major release.
+        DEPRECATED: use :meth:`query(raw=...) <.query>` instead. This method will be removed in the next major release.
 
         Parameters
         ----------
@@ -217,12 +249,10 @@ class SentinelAPI:
     def count(self, area=None, date=None, raw=None, area_relation='Intersects', **keywords):
         """Get the number of products matching a query.
 
-        This is a significantly more efficient alternative to doing len(api.query()),
-        which can take minutes to run for queries matching thousands of products.
+        Accepted parameters are identical to :meth:`SentinelAPI.query()`.
 
-        Parameters
-        ----------
-        Identical to the parameters of api.query().
+        This is a significantly more efficient alternative to doing `len(api.query())`,
+        which can take minutes to run for queries matching thousands of products.
 
         Returns
         -------
@@ -268,7 +298,9 @@ class SentinelAPI:
 
         # load query results
         url = self._format_url(order_by, limit, offset)
-        response = self.session.post(url, {'q': query}, auth=self.session.auth)
+        response = self.session.post(url, {'q': query}, auth=self.session.auth,
+                                     headers={'Content-Type': 'application/x-www-form-urlencoded; charset=UTF-8'},
+                                     timeout=self.timeout)
         _check_scihub_response(response)
 
         # store last status code (for testing)
@@ -348,8 +380,11 @@ class SentinelAPI:
         except ImportError:
             raise ImportError("to_geodataframe requires the optional dependencies GeoPandas and Shapely.")
 
-        df = SentinelAPI.to_dataframe(products)
         crs = {'init': 'epsg:4326'}  # WGS84
+        if len(products) == 0:
+            return gpd.GeoDataFrame(crs=crs)
+
+        df = SentinelAPI.to_dataframe(products)
         geometry = [shapely.wkt.loads(fp) for fp in df['footprint']]
         # remove useless columns
         df.drop(['footprint', 'gmlfootprint'], axis=1, inplace=True)
@@ -361,33 +396,70 @@ class SentinelAPI:
         Returns a dict containing the id, title, size, md5sum, date, footprint and download url
         of the product. The date field corresponds to the Start ContentDate value.
 
-        If ``full`` is set to True, then the full, detailed metadata of the product is returned
-        in addition to the above. For a mapping between the OpenSearch (Solr) and OData
-        attribute names see the following definition files:
-        https://github.com/SentinelDataHub/DataHubSystem/blob/master/addon/sentinel-1/src/main/resources/META-INF/sentinel-1.owl
-        https://github.com/SentinelDataHub/DataHubSystem/blob/master/addon/sentinel-2/src/main/resources/META-INF/sentinel-2.owl
-        https://github.com/SentinelDataHub/DataHubSystem/blob/master/addon/sentinel-3/src/main/resources/META-INF/sentinel-3.owl
+        If `full` is set to True, then the full, detailed metadata of the product is returned
+        in addition to the above.
 
         Parameters
         ----------
         id : string
-            The ID of the product to query
+            The UUID of the product to query
         full : bool
-            Whether to get the full metadata for the Product
+            Whether to get the full metadata for the Product. False by default.
 
         Returns
         -------
         dict[str, Any]
             A dictionary with an item for each metadata attribute
+
+        Notes
+        -----
+        For a full list of mappings between the OpenSearch (Solr) and OData attribute names
+        see the following definition files:
+        https://github.com/SentinelDataHub/DataHubSystem/blob/master/addon/sentinel-1/src/main/resources/META-INF/sentinel-1.owl
+        https://github.com/SentinelDataHub/DataHubSystem/blob/master/addon/sentinel-2/src/main/resources/META-INF/sentinel-2.owl
+        https://github.com/SentinelDataHub/DataHubSystem/blob/master/addon/sentinel-3/src/main/resources/META-INF/sentinel-3.owl
         """
-        url = urljoin(self.api_url, "odata/v1/Products('{}')?$format=json".format(id))
+        url = urljoin(self.api_url, u"odata/v1/Products('{}')?$format=json".format(id))
         if full:
             url += '&$expand=Attributes'
-        response = self.session.get(url, auth=self.session.auth)
+        response = self.session.get(url, auth=self.session.auth,
+                                    timeout=self.timeout)
         _check_scihub_response(response)
         values = _parse_odata_response(response.json()['d'])
         return values
 
+    def _trigger_offline_retrieval(self, url):
+        """ Triggers retrieval of an offline product
+
+        Trying to download an offline product triggers its retrieval from the long term archive.
+        The returned HTTP status code conveys whether this was successful.
+
+        Parameters
+        ----------
+        url : string
+            URL for downloading the product
+
+        Notes
+        -----
+        https://scihub.copernicus.eu/userguide/LongTermArchive
+
+        """
+        with self.session.get(url, auth=self.session.auth, timeout=self.timeout) as r:
+            # check https://scihub.copernicus.eu/userguide/LongTermArchive#HTTP_Status_codes
+            if r.status_code == 202:
+                self.logger.info("Accepted for retrieval")
+            elif r.status_code == 503:
+                self.logger.error("Request not accepted")
+                raise SentinelAPILTAError('Request for retrieval from LTA not accepted', r)
+            elif r.status_code == 403:
+                self.logger.error("Requests exceed user quota")
+                raise SentinelAPILTAError('Requests for retrieval from LTA exceed user quota', r)
+            elif r.status_code == 500:
+                # should not happen
+                self.logger.error("Trying to download an offline product")
+                raise SentinelAPILTAError('Trying to download an offline product', r)
+            return r.status_code
+
     def download(self, id, directory_path='.', checksum=True):
         """Download a product.
 
@@ -429,6 +501,14 @@ class SentinelAPI:
             # We assume that the product has been downloaded and is complete
             return product_info
 
+        # An incomplete download triggers the retrieval from the LTA if the product is not online
+        if not product_info['Online']:
+            self.logger.warning(
+                'Product %s is not online. Triggering retrieval from long term archive.',
+                product_info['id'])
+            self._trigger_offline_retrieval(product_info['url'])
+            return product_info
+
         # Use a temporary file for downloading
         temp_path = path + '.incomplete'
 
@@ -492,10 +572,10 @@ class SentinelAPI:
             Directory where the downloaded files will be downloaded
         max_attempts : int, optional
             Number of allowed retries before giving up downloading a product. Defaults to 10.
-
-        Other Parameters
-        ----------------
-        See download().
+        checksum : bool, optional
+            If True, verify the downloaded files' integrity by checking its MD5 checksum.
+            Throws InvalidChecksumError if the checksum does not match.
+            Defaults to True.
 
         Raises
         ------
@@ -506,6 +586,9 @@ class SentinelAPI:
         dict[string, dict]
             A dictionary containing the return value from download() for each successfully
             downloaded product.
+        dict[string, dict]
+            A dictionary containing the product information for products whose retrieval
+            from the long term archive was successfully triggered.
         set[string]
             The list of products that failed to download.
         """
@@ -525,19 +608,27 @@ class SentinelAPI:
                     last_exception = e
                     self.logger.warning(
                         "Invalid checksum. The downloaded file for '%s' is corrupted.", product_id)
+                except SentinelAPILTAError as e:
+                    last_exception = e
+                    self.logger.exception("There was an error retrieving %s from the LTA", product_id)
+                    break
                 except Exception as e:
                     last_exception = e
                     self.logger.exception("There was an error downloading %s", product_id)
             self.logger.info("%s/%s products downloaded", i + 1, len(product_ids))
         failed = set(products) - set(return_values)
 
+        # split up sucessfully processed products into downloaded and only triggered retrieval from the LTA
+        triggered = OrderedDict([(k, v) for k, v in return_values.items() if v['Online'] is False])
+        downloaded = OrderedDict([(k, v) for k, v in return_values.items() if v['Online'] is True])
+
         if len(failed) == len(product_ids) and last_exception is not None:
             raise last_exception
-        return return_values, failed
+        return downloaded, triggered, failed
 
     @staticmethod
     def get_products_size(products):
-        """Return the total file size in GB of all products in the OpenSearch response"""
+        """Return the total file size in GB of all products in the OpenSearch response."""
         size_total = 0
         for title, props in products.items():
             size_product = props["size"]
@@ -554,9 +645,8 @@ class SentinelAPI:
     def check_query_length(query):
         """Determine whether a query to the OpenSearch API is too long.
 
-        The length of a query string is limited to approximately 3893 characters but
-        any special characters (that is, not alphanumeric or -_.~) are counted twice
-        towards that limit.
+        The length of a query string is limited to approximately 3938 characters but
+        any special characters (that is, not alphanumeric or -_.*) will take up more space.
 
         Parameters
         ----------
@@ -567,23 +657,10 @@ class SentinelAPI:
         -------
         float
             Ratio of the query length to the maximum length
-
-        Notes
-        -----
-        The query size limit arises from a limit on the length of the server's internal query,
-        which looks like
-
-        http://localhost:30333//solr/dhus/select?q=...
-        &wt=xslt&tr=opensearch_atom.xsl&dhusLongName=Sentinels+Scientific+Data+Hub
-        &dhusServer=https%3A%2F%2Fscihub.copernicus.eu%2Fapihub%2F&originalQuery=...
-        &rows=100&start=0&sort=ingestiondate+desc
-
-        This function will estimate the length of the "q" and "originalQuery" parameters to
-        determine whether the query will fail. Their combined length can be at most about
-        7786 bytes.
         """
-        effective_length = len(query) + 2 * len(re.findall('[^-_.* 0-9A-Za-z]', query))
-        return effective_length / 3950
+        # The server uses the Java's URLEncoder implementation internally, which we are replicating here
+        effective_length = len(quote_plus(query, safe="-_.*").replace('~', '%7E'))
+        return effective_length / 3938
 
     def _query_names(self, names):
         """Find products by their names, e.g.
@@ -653,7 +730,7 @@ class SentinelAPI:
         dict[str, list[dict]]
             A dictionary listing the invalid or missing files. The dictionary maps the corrupt
             file paths to a list of OData dictionaries of matching products on the server (as
-            returned by ``SentinelAPI.get_product_odata()``).
+            returned by :meth:`SentinelAPI.get_product_odata()`).
         """
         if not ids and not paths:
             raise ValueError("Must provide either file paths or product IDs and a directory")
@@ -716,7 +793,7 @@ class SentinelAPI:
         return corrupt
 
     def _md5_compare(self, file_path, checksum, block_size=2 ** 13):
-        """Compare a given md5 checksum with one calculated from a file"""
+        """Compare a given MD5 checksum with one calculated from a file."""
         with closing(self._tqdm(desc="MD5 checksumming", total=getsize(file_path), unit="B",
                                 unit_scale=True)) as progress:
             md5 = hashlib.md5()
@@ -738,7 +815,8 @@ class SentinelAPI:
         else:
             already_downloaded_bytes = 0
         downloaded_bytes = 0
-        with closing(session.get(url, stream=True, auth=session.auth, headers=headers)) as r, \
+        with closing(session.get(url, stream=True, auth=session.auth,
+                                 headers=headers, timeout=self.timeout)) as r, \
                 closing(self._tqdm(desc="Downloading", total=file_size, unit="B",
                                    unit_scale=True, initial=already_downloaded_bytes)) as progress:
             _check_scihub_response(r, test_json=False)
@@ -761,6 +839,13 @@ class SentinelAPI:
 
 class SentinelAPIError(Exception):
     """Invalid responses from DataHub.
+
+    Attributes
+    ----------
+    msg: str
+        The error message.
+    response: requests.Response
+        The response from the server as a `requests.Response` object.
     """
 
     def __init__(self, msg=None, response=None):
@@ -772,6 +857,21 @@ class SentinelAPIError(Exception):
             self.response.status_code, self.response.reason,
             ('\n' if '\n' in self.msg else '') + self.msg)
 
+class SentinelAPILTAError(SentinelAPIError):
+    """ Error when retrieving a product from the Long Term Archive
+
+    Attributes
+    ----------
+    msg: str
+        The error message.
+    response: requests.Response
+        The response from the server as a `requests.Response` object.
+    """
+
+    def __init__(self, msg=None, response=None):
+        self.msg = msg
+        self.response = response
+
 
 class InvalidChecksumError(Exception):
     """MD5 checksum of a local file does not match the one from the server.
@@ -810,6 +910,8 @@ def geojson_to_wkt(geojson_obj, feature_number=0, decimals=4):
     """
     if 'coordinates' in geojson_obj:
         geometry = geojson_obj
+    elif 'geometry' in geojson_obj:
+        geometry = geojson_obj['geometry']
     else:
         geometry = geojson_obj['features'][feature_number]['geometry']
 
@@ -819,8 +921,18 @@ def geojson_to_wkt(geojson_obj, feature_number=0, decimals=4):
         else:
             return geometry[:2]
 
+    def check_bounds(geometry):
+        if isinstance(geometry[0], (list, tuple)):
+            return list(map(check_bounds, geometry))
+        else:
+            if geometry[0] > 180 or geometry[0] < -180:
+                raise ValueError('Longitude is out of bounds, check your JSON format or data')
+            if geometry[1] > 90 or geometry[1] < -90:
+                raise ValueError('Latitude is out of bounds, check your JSON format or data')
+
     # Discard z-coordinate, if it exists
     geometry['coordinates'] = ensure_2d(geometry['coordinates'])
+    check_bounds(geometry['coordinates'])
 
     wkt = geomet.wkt.dumps(geometry, decimals=decimals)
     # Strip unnecessary spaces
@@ -829,14 +941,39 @@ def geojson_to_wkt(geojson_obj, feature_number=0, decimals=4):
 
 
 def format_query_date(in_date):
-    """Format a date, datetime or a YYYYMMDD string input as YYYY-MM-DDThh:mm:ssZ
-    or validate a string input as suitable for the full text search interface and return it.
     """
+    Format a date, datetime or a YYYYMMDD string input as YYYY-MM-DDThh:mm:ssZ
+    or validate a date string as suitable for the full text search interface and return it.
+
+    `None` will be converted to '\*', meaning an unlimited date bound in date ranges.
+
+    Parameters
+    ----------
+    in_date : str or datetime or date or None
+        Date to be formatted
+
+    Returns
+    -------
+    str
+        Formatted string
+
+    Raises
+    ------
+    ValueError
+        If the input date type is incorrect or passed date string is invalid
+    """
+    if in_date is None:
+        return '*'
     if isinstance(in_date, (datetime, date)):
         return in_date.strftime('%Y-%m-%dT%H:%M:%SZ')
     elif not isinstance(in_date, string_types):
         raise ValueError('Expected a string or a datetime object. Received {}.'.format(in_date))
 
+    in_date = in_date.strip()
+    if in_date == '*':
+        # '*' can be used for one-sided range queries e.g. ingestiondate:[* TO NOW-1YEAR]
+        return in_date
+
     # Reference: https://cwiki.apache.org/confluence/display/solr/Working+with+Dates
 
     # ISO-8601 date or NOW
@@ -987,7 +1124,10 @@ def _parse_odata_response(product):
         product['Checksum']['Algorithm'].lower(): product['Checksum']['Value'],
         'date': _parse_odata_timestamp(product['ContentDate']['Start']),
         'footprint': _parse_gml_footprint(product["ContentGeometry"]),
-        'url': product['__metadata']['media_src']
+        'url': product['__metadata']['media_src'],
+        'Online': product['Online'],
+        'Creation Date': _parse_odata_timestamp(product['CreationDate']),
+        'Ingestion Date': _parse_odata_timestamp(product['IngestionDate']),
     }
     # Parse the extended metadata, if provided
     converters = [int, float, _parse_iso_date]
